﻿-- ********************************************************************************
-- Broker Statistics
-- A Data Broker display for combat statistics.
-- By: Shenton
--
-- Core.lua
-- Ace3 libs, core functions, Ace3 init.
-- ********************************************************************************

-- Ace libs (<3)
local A = LibStub("AceAddon-3.0"):NewAddon("BrokerStats", "AceConsole-3.0", "AceTimer-3.0", "AceComm-3.0", "AceEvent-3.0");
local L = LibStub("AceLocale-3.0"):GetLocale("BrokerStats");

-- Addon global
_G["DataBrokerStatisticsGlobal"] = A;

-- LUA globals
local mfloor, pairs, ipairs, loadstring = math.floor, pairs, ipairs, loadstring;
local smatch, sgsub, tostring, type = string.match, string.gsub, tostring, type;
local tinsert, tconcat, format, tsort = table.insert, table.concat, string.format, table.sort;

-- WoW globals
local strtrim = strtrim;

-- ********************************************************************************
-- Functions
-- ********************************************************************************

--- Send a message to the chat frame with the addon name colored
-- @param text The message to display
-- @param level Verbosity level (0 to 2)
-- @param color Bool, if true will color in red
function A:Message(text, level, color)
    if ( level <= A.db.profile.verbosity ) then
        if ( color ) then
            color = A.colorRed;
        else
            color = A.colorGreen;
        end

        DEFAULT_CHAT_FRAME:AddMessage(color..L["Broker Statistics"]..": "..A.colorReset..text);
    end
end

--- Handle the slash command
function A:SlashCommand()
    LibStub("AceConfigDialog-3.0"):Open("BrokerStatsConfig");
end

--- Round function (http://lua-users.org/wiki/SimpleRound)
-- @param num The number to round
-- @param idp Decimal places
function A:Round(num, idp)
    if ( type(num) == "number" ) then
        local mult = 10^(idp or 0);
        return mfloor(num * mult + 0.5) / mult;
    else
        A:Message(L["Attempt to perform arithmetic on a non number value, you are probably trying to round a string."], 1, 1);
        return num;
    end
end

-- Convert table to string (http://lua-users.org/wiki/TableUtils)
function A:TableValToStr(v)
    if ( "string" == type(v) ) then
        v = sgsub(v, "\n", "\\n");
        if ( smatch(sgsub(v, "[^'\"]", ""), '^"+$') ) then return "'"..v.."'"; end
        return '"'..sgsub(v, '"', '\\"')..'"';
    else
        return "table" == type(v) and A:TableTostring(v) or tostring(v);
    end
end
function A:TableKeyToStr(k)
    if ( "string" == type(k) and smatch(k, "^[_%a][_%a%d]*$") ) then return k; end

    return "["..A:TableValToStr( k ) .. "]";
end
function A:TableTostring(tbl)
    local result, done = {}, {};

    for k,v in ipairs(tbl) do
        tinsert(result, A:TableValToStr(v));
        done[k] = true;
    end

    for k, v in pairs(tbl) do
        if ( not done[k] ) then tinsert(result, A:TableKeyToStr(k).."="..A:TableValToStr(v)); end
    end

    return "{"..tconcat(result, ",").."}";
end

--- Pairs function with alphabetic sort
function A:PairsByKeys(t, f)
    local a, i = {}, 0;

    for n in pairs(t) do a[#a+1] = n; end
    tsort(a, f);

    local iter = function()
        i = i + 1;
        if ( not a[i] ) then
            return nil;
        else
            return a[i], t[a[i]];
        end
    end

    return iter;
end

--- Remove unused entries from the database
function A:DelUnusedEntry()
    for k in pairs(A.db.profile.buildInStats) do
        for kk,vv in pairs(A.db.profile.buildInStats[k]) do
            if ( vv.dbObject ) then
                if ( not A.db.profile.dataBroker.customDiplays[vv.dbObject] ) then
                    A.db.profile.buildInStats[k][kk].dbObject = "Default";
                end
            end
        end
    end

    for k in pairs(A.db.profile.custom) do
        for kk,vv in pairs(A.db.profile.custom[k]) do
            if ( vv.dbObject ) then
                if ( not A.db.profile.dataBroker.customDiplays[vv.dbObject] ) then
                    A.db.profile.custom[k][kk].dbObject = "Default";
                end
            end
        end
    end
end

-- Return if there is a least one custom stat
function A:GotCustom()
    for _,v in pairs(A.db.global.custom) do
        if ( v.calc and v.calc ~= "" ) then return 1; end
    end

    return nil;
end

-- Return if there is a least one custom ldb object
function A:GotCustomLdb()
    for k in pairs(A.db.profile.dataBroker.customDiplays) do
        if ( k ) then return 1; end
    end

    return nil;
end

--- Create the select table for ldb objects
local out;
function A:CreateDbObjectsSelect()
    out =
    {
        [L["Default"]] = L["Default"];
    };

    for k in pairs(A.db.profile.dataBroker.customDiplays) do
        out[k] = k;
    end

    return out;
end

--- Set the current spec var to the player current spec
function A:SetSpecVar()
    local spec = GetActiveTalentGroup(false, false);

    if ( spec == 1 ) then
        A.playerSpec = "primary";
    else
        A.playerSpec = "secondary";
    end
end

--- Pause toggle
function A:PauseToggle()
    if ( A.paused ) then
        A.mainTimer = A:ScheduleRepeatingTimer("UpdateText", A.db.profile.time);
        A.paused = nil;
    else
        A:CancelTimer(A.mainTimer, 1);
        A:SetTextOnLdbObjects(A.colorRed..L["Paused"]);
        A.paused = 1;
    end
end

--- Count the number of activated statistics and ldb objects
-- @return the number of stats and objects activated and theyr names
function A:CountActivated()
    local stats, countStats, objects, countObjects = {}, 0, {}, 0;

    for k,v in pairs(A.db.profile.buildInStats[A.playerSpec]) do
        if ( v.enabled ) then
            stats[#stats+1] = k;
            countStats = countStats + 1;
        end
    end

    for k,v in pairs(A.db.profile.custom[A.playerSpec]) do
        if ( v.enabled ) then
            stats[#stats+1] = k;
            countStats = countStats + 1;
        end
    end

    for k,v in pairs(A.db.profile.dataBroker.customDiplays) do
        if ( v.enabled ) then
            objects[#objects+1] = k;
            countObjects = countObjects + 1;
        end
    end

    return stats, countStats, objects, countObjects;
end

--- Display a string on all enabled ldb objects
function A:SetTextOnLdbObjects(text)
    A.ldbObject.text = text;

    for k,v in pairs(A.db.profile.dataBroker.customDiplays) do
        if ( v.enabled ) then A["ldbObject"..k].text = text; end
    end
end

--- Refresh the main timer to the current delay
function A:RefreshTimer()
    A.db.profile.time = A:Round(A.db.profile.time, 1);
    A:CancelTimer(A.mainTimer, 1);
    A.mainTimer = A:ScheduleRepeatingTimer("UpdateText", A.db.profile.time);
end

--- Create the others Data Broker objects if needed
function A:CreateDataBrokerOjbects()
    for k,v in pairs(A.db.profile.dataBroker.customDiplays) do
        if ( v.enabled ) then
            if ( LibStub("LibDataBroker-1.1"):GetDataObjectByName("Broker Stats - "..k) ) then
                A["ldbObject"..k] = LibStub("LibDataBroker-1.1"):GetDataObjectByName("Broker Stats - "..k);
            else
                A["ldbObject"..k] = LibStub("LibDataBroker-1.1"):NewDataObject("Broker Stats - "..k,
                {
                    type = "data source",
                    text = L["Not available"],
                    label = L["Broker Statistics - "..k],
                    icon = "Interface\\ICONS\\Spell_Magic_PolymorphRabbit",
                    tocname = "Broker_Stats",
                    OnClick = function(self, button)
                        if (button == "LeftButton") then
                            A:PauseToggle();
                        elseif ( button == "RightButton" ) then
                            A:DropDownMenu(self, k);
                        elseif ( button == "MiddleButton" ) then
                            LibStub("AceConfigDialog-3.0"):Open("BrokerStatsConfig");
                        end
                    end,
                    OnTooltipShow = function(tooltip)
                        local countStats, countObjects = A:CountActivated();

                        tooltip:AddDoubleLine(A.colorWhite..L["Broker Statistics"], A.colorGreen.."v"..A.version);
                        tooltip:AddLine(" ");
                        tooltip:AddLine(L["%d statistics displayed, with %d Data Broker custom objects."]:format(countStats, countObjects));
                        tooltip:AddLine(" ");
                        tooltip:AddLine(L["|cFFC79C6ELeft-Click: |cFF33FF99Pause the addon update timer\n|cFFC79C6ERight-Click: |cFF33FF99Open the quick access menu\n|cFFC79C6EMiddle-Click: |cFF33FF99Open the configuration panel"]);
                    end
                });
            end
        end
    end
end

--- Update text on the data broker display
local updateText;
function A:UpdateText()
    updateText =
    {
        Default = ""
    };

    for k,v in pairs(A.db.profile.dataBroker.customDiplays) do
        if ( v.enabled ) then updateText[k] = ""; end
    end

    for k,v in pairs(A.db.profile.buildInStats[A.playerSpec]) do
        if ( v.enabled ) then
            local opt = A.buildInStats[k];
            local pre, post, calc, dbObject = "", "", nil, v.dbObject;

            if ( not dbObject ) then dbObject = "Default"; end

            if ( updateText[dbObject] ) then
                if ( opt.pre and A.db.profile.displayPre ) then pre = opt.pre; end
                if ( opt.post and A.db.profile.displayPost ) then post = opt.post; end
                if ( opt.calc ) then calc = loadstring(opt.calc); end

                if ( calc ) then
                    if ( opt.multiple ) then
                        local values = { calc() };
                        calc = "";

                        for _,vv in pairs(values) do
                            if ( opt.round ) then
                                calc = calc..A:Round(vv, opt.round)..post..opt.separator;
                            else
                                calc = calc..vv..post..opt.separator;
                            end
                        end

                        calc = strtrim(calc, opt.separator);
                        post = "";
                    else
                        if ( opt.round ) then
                            calc = A:Round(calc(), opt.round);
                        else
                            calc = calc();
                        end
                    end

                    updateText[dbObject] = updateText[dbObject].." ".."|cff5577bb"..pre.."|r"..calc..post;
                end
            end
        end
    end

    for k,v in pairs(A.db.profile.custom[A.playerSpec]) do
        if ( v.enabled ) then
            local opt = A.db.global.custom[k];
            local pre, post, calc, dbObject = "", "", nil, v.dbObject;

            if ( not dbObject ) then dbObject = "Default"; end

            if ( updateText[dbObject] ) then
                if ( opt.pre and A.db.profile.displayPre ) then pre = opt.pre; end
                if ( opt.post and A.db.profile.displayPost ) then post = opt.post; end
                if ( opt.calc ) then calc = loadstring(opt.calc); end

                local ok, errorString = pcall(calc);

                if ( ok ) then
                    if ( opt.multiple ) then
                        local values = { calc() };
                        calc = "";

                        for _,vv in pairs(values) do
                            if ( opt.round ) then
                                calc = calc..A:Round(vv, opt.round)..post..opt.separator;
                            else
                                calc = calc..vv..post..opt.separator;
                            end
                        end

                        calc = strtrim(calc, opt.separator);
                        post = "";
                    else
                        if ( opt.round ) then
                            calc = A:Round(calc(), opt.round);
                        else
                            calc = calc();
                        end
                    end

                    if ( calc ) then
                        updateText[dbObject] = updateText[dbObject].." ".."|cff5577bb"..pre.."|r"..calc..post;
                    else
                        A:Message(L["There is an error in %s code. Error: %s"]:format(k, L["Returned value is nil."]), 1, 1);
                    end
                else
                    A:Message(L["There is an error in %s code. Error: %s"]:format(k, errorString), 1, 1);
                end
            end
        end
    end

    for k,v in pairs(updateText) do
        if ( v == "" ) then v = L["Not available"]; end

        if ( k == "Default" ) then
            A.ldbObject.text = strtrim(v);
        else
            A["ldbObject"..k].text = strtrim(v);
        end
    end
end

-- ********************************************************************************
-- Addon Communication
-- ********************************************************************************

--- Called when receiving an addon communication message
function A:AddonComm(...)
    local prefix, customText, method, who, customName;

    prefix, customText, method, who = ...;

    if ( prefix ~= "BrokerStatsComm" ) then return; end
    if ( who == UnitName("player") ) then return; end

    if ( (method == "WHISPER" and A.db.profile.shareWhisper) or
    (method == "RAID" and A.db.profile.shareRaid) or
    (method == "GUILD" and A.db.profile.shareGuild) or
    (method == "BATTLEGROUND" and A.db.profile.shareBg) ) then
        A:Message(L["%s tried to send you a custom stat (%s) with method %s."]:format(who, customName, method), 2, 1);
        return;
    end

    customName = smatch(customText, "^(.+){");
    customText = smatch(customText, "{(.+)}");
    customText = "{"..customText.."}";

    if ( A.db.global.custom[customName] ) then
        local pop = StaticPopup_Show("BrokerStatsReceiveCustomNameExists", customName, who);
        if (pop) then pop.customText = customText; end
    else
        local pop = StaticPopup_Show("BrokerStatsReceiveCustom", customName, who);
        if (pop) then
            pop.customText = customText;
            pop.customName = customName;
        end
    end
end

--- Send an addon communication message with a custom stat
-- @param target The message target
-- @param customStat The custom stat name
function A:SendStat(target, customStat)
    if ( not target or not customStat ) then return; end

    target = strtrim(target);
    if ( target == "" ) then return; end

    local textMessage = A:TableTostring(A.db.global.custom[customStat]);

    if ( target == "RAID" or target == "GUILD" or target == "BATTLEGROUND" ) then
        A:SendCommMessage("BrokerStatsComm", customStat..textMessage, target);
    else
        A:SendCommMessage("BrokerStatsComm", customStat..textMessage, "WHISPER", target);
    end
end

-- ********************************************************************************
-- Callbacks
-- ********************************************************************************

--- Called when an action is made on a configuration profile
function A:RefreshConfig()
    A:RefreshTimer();
    A:CreateDataBrokerOjbects();
end

--- Callback for event ACTIVE_TALENT_GROUP_CHANGED
function A:ACTIVE_TALENT_GROUP_CHANGED()
    A:SetSpecVar();
    LibStub("AceConfigRegistry-3.0"):NotifyChange("BrokerStatsConfig");
end

-- ********************************************************************************
-- Main
-- ********************************************************************************

--- AceAddon callback
-- Called after the addon is fully loaded
function A:OnInitialize()
    -- Database
    A.db = LibStub("AceDB-3.0"):New("BrokerStatsDB", A:AceDefaultDB());

    -- Config settings refresh
    A.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig");
    A.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig");
    A.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig");

    -- Slash command
    A:RegisterChatCommand("brokerstats", "SlashCommand");
    A:RegisterChatCommand("stats", "SlashCommand");
    A:RegisterChatCommand("bs", "SlashCommand");
end

--- AceAddon callback
-- Called during the PLAYER_LOGIN event
function A:OnEnable()
    -- Config
    LibStub("AceConfig-3.0"):RegisterOptionsTable("BrokerStatsConfig", A.AceConfig);
    LibStub("AceConfigDialog-3.0"):SetDefaultSize("BrokerStatsConfig", 800, 500);
    LibStub("AceConfigDialog-3.0"):AddToBlizOptions("BrokerStatsConfig", L["Broker Statistics"]);

    -- LDB
    if ( LibStub("LibDataBroker-1.1"):GetDataObjectByName("Broker Stats") ) then
        A.ldbObject = LibStub("LibDataBroker-1.1"):GetDataObjectByName("Broker Stats");
    else
        A.ldbObject = LibStub("LibDataBroker-1.1"):NewDataObject("Broker Stats",
        {
            type = "data source",
            text = L["Not available"],
            label = L["Broker Statistics"],
            icon = "Interface\\ICONS\\Spell_Magic_PolymorphRabbit",
            tocname = "Broker_Stats",
            OnClick = function(self, button)
                if (button == "LeftButton") then
                    A:PauseToggle();
                elseif ( button == "RightButton" ) then
                    A:DropDownMenu(self, "Default");
                elseif ( button == "MiddleButton" ) then
                    LibStub("AceConfigDialog-3.0"):Open("BrokerStatsConfig");
                end
            end,
            OnTooltipShow = function(tooltip)
                local statsNames, countStats, objectsName, countObjects = A:CountActivated();

                objectsName[#objectsName+1] = L["Default"];

                statsNames = tconcat(statsNames, ", ");
                objectsName = tconcat(objectsName, ", ");

                tooltip:AddDoubleLine(A.colorWhite..L["Broker Statistics"], A.colorGreen.."v"..A.version);
                tooltip:AddLine(" ");
                tooltip:AddLine(L["Statistics displayed: %d"]:format(countStats));
                tooltip:AddLine(statsNames);
                tooltip:AddLine(" ");
                tooltip:AddLine(L["Data broker objects used: %d"]:format(countObjects + 1));
                tooltip:AddLine(objectsName);
                tooltip:AddLine(" ");
                tooltip:AddLine(L["|cFFC79C6ELeft-Click: |cFF33FF99Pause the addon update timer\n|cFFC79C6ERight-Click: |cFF33FF99Open the quick access menu\n|cFFC79C6EMiddle-Click: |cFF33FF99Open the configuration panel"]);
            end
        });
    end

    -- DropDownMenu frame
    A.dropDownMenuFrame = CreateFrame("Frame", "BrokerStatisticsMenuFrame");
    A.dropDownMenuFrame.displayMode = "MENU";

    -- Delete unused database entry
    A:DelUnusedEntry();

    -- Events
    A:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED");

    -- Create Custom ldb objects
    A:CreateDataBrokerOjbects();

    -- Register for addon communication
    A:RegisterComm("BrokerStatsComm", "AddonComm");

    -- Player spec
    A:SetSpecVar();

    -- Main timer
    A.db.profile.time = A:Round(A.db.profile.time, 1);
    A.mainTimer = A:ScheduleRepeatingTimer("UpdateText", A.db.profile.time);
end

--- AceAddon callback
-- Called when the addon is manually disabled
function A:OnDisable()
    A:CancelTimer(A.mainTimer, 1);
    A:UnregisterEvent("ACTIVE_TALENT_GROUP_CHANGED");
end
